/*
 * Sample program for offscreen bitmaps, using v_open_bm
 */

#include <stddef.h>
#ifdef __GNUC__
#include <gemx.h>
#include <stdint.h>
#include <mint/osbind.h>
#ifndef _WORD
#define _WORD short
#endif
#else
#include <tos.h>
#include <vdi.h>
#include <aes.h>
#endif

/*----------------------------------------------------------------------------------------*/
/* global variables                                                                       */
/*----------------------------------------------------------------------------------------*/
static _WORD app_id;
static _WORD aes_handle;

/*----------------------------------------------------------------------------------------*/
/* Open virtual workstation                                                               */
/* Result:              VDI-handle, or 0 for error                                        */
/* work_out:            device information                                                */
/*----------------------------------------------------------------------------------------*/
static _WORD open_screen_wk(_WORD aes_handle, _WORD *work_out)
{
	_WORD work_in[11];
	_WORD handle;
	_WORD i;

	for (i = 0; i < 10; i++)
		work_in[i] = 1;

	work_in[10] = 2;					/* use raster cordinates */
	handle = aes_handle;

	v_opnvwk(work_in, &handle, work_out);
	return handle;
}

/*----------------------------------------------------------------------------------------*/
/* Initialize GCBITMAP-structure without color table                                      */
/* Resultat:            -                                                                 */
/* bm:                  pointer to structure                                              */
/* addr:                pointer to framebuffer, or NULL                                   */
/* px_format:           pixel format                                                      */
/* width:               width in pixel                                                    */
/* height:              height in pixel                                                   */
/*----------------------------------------------------------------------------------------*/
static void init_gcbitmap(GCBITMAP *bm, unsigned char *addr, int32_t px_format, int32_t width, int32_t height)
{
	bm->magic = CBITMAP_MAGIC; 			/* magic */
	bm->length = sizeof(GCBITMAP);		/* size of structure */
	bm->format = 0;						/* Format 0 */
	bm->reserved = 0;

	if (addr)							/* do we have memory? */
	{
		bm->addr = addr;
		bm->width = (width * (px_format & PX_BITS)) >> 3;	/* line width in bytes */
	} else
	{
		bm->addr = 0;
		bm->width = 0;
	}

	bm->bits = px_format & PX_BITS;
	bm->px_format = px_format;

	bm->xmin = 0;
	bm->ymin = 0;
	bm->xmax = width;
	bm->ymax = height;

	bm->ctab = 0;
	bm->itab = 0;
	bm->color_space = 0;
	bm->reserved1 = 0;
}

/*----------------------------------------------------------------------------------------*/
/* Create offscreen bitmap with own color table.                                          */
/* Changes to colors on the screen have no influence to drawing the bitmap, because       */
/* vr_transfer_bits() will transform the colors                                           */
/*----------------------------------------------------------------------------------------*/
static int dont_share_colors(void)
{
	_WORD screen_out[57];
	_WORD screen_handle;
	int ret = 0;

	wind_update(BEG_UPDATE);			/* lock screen */
	wind_update(BEG_MCTRL); 			/* get keyboard and mouse control */

	screen_handle = open_screen_wk(aes_handle, screen_out);	/* get VDI handle */

	if (screen_handle)
	{
		GRECT size;
		GCBITMAP bm;
		_WORD bm_handle;

		size.g_x = 0;
		size.g_x = 0;
		size.g_w = screen_out[0] + 1;
		size.g_h = screen_out[1] + 1;
		form_dial_grect(FMD_START, &size, &size);

		/* initialize GCBITMAP, using dimensions of 0 to let v_open_bm() */
		/* use the values from <screen_handle> */
		init_gcbitmap(&bm, 0, 0, 0, 0);

		/* Create bitmap with screen dimensions and format. */
		/* The bitmap will have its own color table! */
		bm_handle = v_open_bm(screen_handle, &bm, 0, 0, 0, 0);

		if (bm_handle) 				/* bitmap allocated? */
		{
			COLOR_ENTRY blue_color = { { 0, 0x0000, 0x0000, 0xffff } };
			COLOR_ENTRY save_color;
			int32_t save_color_space;
			_WORD ctab_idx;
			_WORD pxy[4];

			/* draw a red rectangle to the screen */
			pxy[0] = 50;
			pxy[1] = 50;
			pxy[2] = 100;
			pxy[3] = 100;
			vsf_color(screen_handle, G_RED);
			vsf_interior(screen_handle, FIS_SOLID);
			vr_recfl(screen_handle, pxy);

			v_gtext(screen_handle, 0, 32, "Red rectangle drawn. Press any key.");
			evnt_keybd();

			/* draw a red circle to the bitmap */
			vsf_color(bm_handle, G_RED);
			vsf_interior(bm_handle, FIS_SOLID);
			v_circle(bm_handle, 200, 200, 100);

			/* draw bitmap transparent to the screen */
			pxy[0] = 0;
			pxy[1] = 0;
			pxy[2] = screen_out[0];
			pxy[3] = screen_out[1];

			vr_transfer_bits(screen_handle, &bm, 0, pxy, pxy, T_TRANSPARENT);

			/* print message and wait for keypress */
			v_gtext(screen_handle, 0, 32, "Bitmap was drawn. Press any key.");
			evnt_keybd();

			/* change color on the screen */
			ctab_idx = v_ctab_vdi2idx(screen_handle, G_RED); /* get color index of red */
			save_color_space = vq_ctab_entry(screen_handle, ctab_idx, &save_color); /* query color values */
			vs_ctab_entry(screen_handle, ctab_idx, CSPACE_RGB, &blue_color);	/* change entry to blue; for up to 256 colors the CLUT will be updated */

			/* print message and wait for keypress */
			v_gtext(screen_handle, 0, 32, "Color was changed from red to blue. Press any key.");
			evnt_keybd();

			/* draw bitmap again; colors will be converted and circle will remain red */
			vr_transfer_bits(screen_handle, &bm, 0, pxy, pxy, T_TRANSPARENT);

			/* print message and wait for keypress */
			v_gtext(screen_handle, 0, 32, "Bitmap was drawn again; color is red again. Press any key.");
			evnt_keybd();

			/* set color to saved values */
			vs_ctab_entry(screen_handle, ctab_idx, save_color_space, &save_color);

			/* print message and wait for keypress */
			v_gtext(screen_handle, 0, 32, "Color is red again. Press any key.");
			evnt_keybd();

			v_clsbm(bm_handle);		/* free bitmap */
			ret = 1;
		} else
		{
			form_alert(1, "[1][Cannot allocate bitmap][ Cancel ]");
		}

		form_dial_grect(FMD_FINISH, &size, &size);	/* let AES redraw screen */
		v_clsvwk(screen_handle);
	} else
	{
		form_alert(1, "[1][Cannot open screen][ Cancel ]");
	}
	wind_update(END_MCTRL);
	wind_update(END_UPDATE);

	return ret;
}

/*----------------------------------------------------------------------------------------*/
/* Create offscreen bitmap with a shared color table.                                     */
/* Drawing operation to screen and bitmap will have the same results, since identical     */
/* pixel values are used.                                                                 */
/*----------------------------------------------------------------------------------------*/
static int share_colors(void)
{
	_WORD screen_out[57];
	_WORD screen_handle;
	int ret = 0;

	wind_update(BEG_UPDATE);			/* lock screen */
	wind_update(BEG_MCTRL); 			/* get keyboard and mouse control */

	screen_handle = open_screen_wk(aes_handle, screen_out);	/* get VDI handle */

	if (screen_handle)
	{
		GRECT size;
		GCBITMAP bm;
		_WORD bm_handle;

		size.g_x = 0;
		size.g_x = 0;
		size.g_w = screen_out[0] + 1;
		size.g_h = screen_out[1] + 1;
		form_dial_grect(FMD_START, &size, &size);

		/* initialize GCBITMAP, using dimensions of 0 to let v_open_bm() */
		/* use the values from <screen_handle> */
		init_gcbitmap(&bm, 0, 0, 0, 0);

		/* Create bitmap with screen dimensions and format. */
		/* The bitmap will have a shared color table! */
		bm_handle = v_open_bm(screen_handle, &bm, 1, 0, 0, 0);

		if (bm_handle) 				/* bitmap allocated? */
		{
			COLOR_ENTRY blue_color = { { 0, 0x0000, 0x0000, 0xffff } };
			COLOR_ENTRY save_color;
			int32_t save_color_space;
			_WORD ctab_idx;
			_WORD pxy[4];

			/* draw a red rectangle to the screen */
			pxy[0] = 50;
			pxy[1] = 50;
			pxy[2] = 100;
			pxy[3] = 100;
			vsf_color(screen_handle, G_RED);
			vsf_interior(screen_handle, FIS_SOLID);
			vr_recfl(screen_handle, pxy);

			v_gtext(screen_handle, 0, 32, "Red rectangle drawn. Press any key.");
			evnt_keybd();

			/* draw a red circle to the bitmap */
			vsf_color(bm_handle, G_RED);
			vsf_interior(bm_handle, FIS_SOLID);
			v_circle(bm_handle, 200, 200, 100);

			/* draw bitmap transparent to the screen */
			pxy[0] = 0;
			pxy[1] = 0;
			pxy[2] = screen_out[0];
			pxy[3] = screen_out[1];

			vr_transfer_bits(screen_handle, &bm, 0, pxy, pxy, T_TRANSPARENT);

			/* print message and wait for keypress */
			v_gtext(screen_handle, 0, 32, "Bitmap was drawn. Press any key.");
			evnt_keybd();

			/* change color on the screen */
			ctab_idx = v_ctab_vdi2idx(screen_handle, G_RED); /* get color index of red */
			save_color_space = vq_ctab_entry(screen_handle, ctab_idx, &save_color); /* query color values */
			vs_ctab_entry(screen_handle, ctab_idx, CSPACE_RGB, &blue_color);	/* change entry to blue; for up to 256 colors the CLUT will be updated */

			/* print message and wait for keypress */
			v_gtext(screen_handle, 0, 32, "Color was changed from red to blue. Press any key.");
			evnt_keybd();

			/* draw bitmap again; colors will not be converted and circle will be drawn in blue */
			vr_transfer_bits(screen_handle, &bm, 0, pxy, pxy, T_TRANSPARENT);

			/* print message and wait for keypress */
			v_gtext(screen_handle, 0, 32, "Bitmap was drawn again; color is blue. Press any key.");
			evnt_keybd();

			/* set color to saved values */
			vs_ctab_entry(screen_handle, ctab_idx, save_color_space, &save_color);

			/* print message and wait for keypress */
			v_gtext(screen_handle, 0, 32, "Color is red again. Press any key.");
			evnt_keybd();

			v_clsbm(bm_handle);		/* free bitmap */
			ret = 1;
		} else
		{
			form_alert(1, "[1][Cannot allocate bitmap][ Cancel ]");
		}

		form_dial_grect(FMD_FINISH, &size, &size);	/* let AES redraw screen */
		v_clsvwk(screen_handle);
	} else
	{
		form_alert(1, "[1][Cannot open screen][ Cancel ]");
	}
	wind_update(END_MCTRL);
	wind_update(END_UPDATE);

	return ret;
}

/*----------------------------------------------------------------------------------------*/
/* Sample program for offscreen bitmaps, using v_open_bm                                  */
/*----------------------------------------------------------------------------------------*/
int main(void)
{
	app_id = appl_init();				/* initialise AES */

	if (app_id >= 0)
	{
		_WORD dummy;

		aes_handle = graf_handle(&dummy, &dummy, &dummy, &dummy);
		graf_mouse(ARROW, NULL);
		graf_mouse(M_OFF, NULL);		/* turn off mouse */

		if (share_colors())				/* draw to bitmap, using shared colors */
		{
			evnt_timer(500);			/* short wait, to let AES redraw screen */
			dont_share_colors(); 		/* draw to bitmap, using own colors */
		}

		graf_mouse(M_ON, NULL);			/* turn mouse on again */
		appl_exit();
	}

	return 0;
}
