/*
 * Sample program for offscreen bitmaps, using v_opnbm
 */

#include <stddef.h>
#ifdef __GNUC__
#include <gemx.h>
#include <stdint.h>
#include <mint/osbind.h>
#ifndef _WORD
#define _WORD short
#endif
#else
#include <tos.h>
#include <vdi.h>
#include <aes.h>
#endif

/*----------------------------------------------------------------------------------------*/
/* global variables                                                                       */
/*----------------------------------------------------------------------------------------*/
static _WORD app_id;
static _WORD aes_handle;

/*----------------------------------------------------------------------------------------*/
/* Open virtual workstation                                                               */
/* Result:              VDI-handle, or 0 for error                                        */
/* work_out:            device information                                                */
/*----------------------------------------------------------------------------------------*/
static _WORD open_screen_wk(_WORD aes_handle, _WORD *work_out)
{
	_WORD work_in[11];
	_WORD handle;
	_WORD i;

	for (i = 0; i < 10; i++)
		work_in[i] = 1;
	work_in[10] = 2;					/* use raster cordinates */

	handle = aes_handle;
	v_opnvwk(work_in, &handle, work_out);
	return handle;
}


static _WORD open_bitmap(MFDB *bm, unsigned char *addr, _WORD width, _WORD height, _WORD planes)
{
	_WORD handle;
	_WORD work_in[20];
	_WORD work_out[57];
	_WORD i;
	
	bm->fd_addr = addr;
	bm->fd_w = width;
	bm->fd_h = height;
	bm->fd_nplanes = planes;
	bm->fd_wdwidth = (width + 15) >> 4;
	bm->fd_stand = 0;
	bm->fd_r1 = 0;
	bm->fd_r2 = 0;
	bm->fd_r3 = 0;

	for (i = 0; i < 10; i++)
		work_in[i] = 1;
	work_in[10] = 2;					/* use raster cordinates */
	work_in[11] = width - 1;
	work_in[12] = height - 1;
	work_in[13] = 0;
	work_in[14] = 0;
	work_in[15] = 0;
	work_in[16] = 0;
	work_in[17] = planes;
	work_in[18] = 0;
	work_in[19] = 1;

	handle = aes_handle;
	v_opnbm(work_in, bm, &handle, work_out);

	return handle;
}


static int draw_bitmap(void)
{
	_WORD screen_out[57];
	_WORD screen_handle;
	int ret = 0;

	wind_update(BEG_UPDATE);			/* lock screen */
	wind_update(BEG_MCTRL); 			/* get keyboard and mouse control */

	screen_handle = open_screen_wk(aes_handle, screen_out);	/* get VDI handle */

	if (screen_handle)
	{
		GRECT size;
		MFDB bm;
		MFDB screen;
		_WORD screen_width, screen_height;
		_WORD bm_handle;

		screen_width = screen_out[0] + 1;
		screen_height = screen_out[1] + 1;
		size.g_x = 0;
		size.g_x = 0;
		size.g_w = screen_width;
		size.g_h = screen_height;
		form_dial_grect(FMD_START, &size, &size);

		/* create bitmap with screen dimensions and format */
		bm_handle = open_bitmap(&bm, NULL, screen_width, screen_height, 0);

		if (bm_handle) 				/* bitmap allocated? */
		{
			_WORD pxy[8];

			/* set clipping for screen */
			pxy[0] = 0;
			pxy[1] = 0;
			pxy[2] = screen_out[0];
			pxy[3] = screen_out[1];
			vs_clip(screen_handle, 1, pxy);
			
			/* set clipping for bitmap */
			pxy[0] = 0;
			pxy[1] = 0;
			pxy[2] = bm.fd_w - 1;
			pxy[3] = bm.fd_h - 1;
			vs_clip(bm_handle, 1, pxy);
			
			/* draw a red rectangle to the screen */
			pxy[0] = 50;
			pxy[1] = 50;
			pxy[2] = 100;
			pxy[3] = 100;
			vsf_color(screen_handle, G_RED);
			vsf_interior(screen_handle, FIS_SOLID);
			vr_recfl(screen_handle, pxy);

			v_gtext(screen_handle, 0, 32, "Red rectangle drawn. Press any key.");
			evnt_keybd();

			/* draw a blue circle to the bitmap */
			vsf_color(bm_handle, G_BLUE);
			vsf_interior(bm_handle, FIS_SOLID);
 			v_ellipse(bm_handle, bm.fd_w / 2, bm.fd_h / 2, bm.fd_w / 4, bm.fd_h / 4);

			/* copy bitmap to screen */
			pxy[0] = 0;
			pxy[1] = 0;
			pxy[2] = screen_out[0];
			pxy[3] = screen_out[1];
			pxy[4] = 0;
			pxy[5] = 0;
			pxy[6] = screen_out[0];
			pxy[7] = screen_out[1];
			screen.fd_addr = 0;
			/* yes, strangely, we have to use screen_handle here */
			vro_cpyfm(screen_handle, S_ONLY, pxy, &bm, &screen);
			
			/* print message and wait for keypress */
			v_gtext(screen_handle, 0, 32, "Bitmap was drawn. Press any key.");
			evnt_keybd();

			v_clsbm(bm_handle);		/* free bitmap */
			ret = 1;
		} else
		{
			form_alert(1, "[1][Cannot allocate bitmap][ Cancel ]");
		}

		form_dial_grect(FMD_FINISH, &size, &size);	/* let AES redraw screen */
		v_clsvwk(screen_handle);
	} else
	{
		form_alert(1, "[1][Cannot open screen][ Cancel ]");
	}
	wind_update(END_MCTRL);
	wind_update(END_UPDATE);

	return ret;
}

/*----------------------------------------------------------------------------------------*/
/* Sample program for offscreen bitmaps, using v_opnbm                                    */
/*----------------------------------------------------------------------------------------*/
int main(void)
{
	app_id = appl_init();				/* initialise AES */

	if (app_id >= 0)
	{
		_WORD dummy;

		aes_handle = graf_handle(&dummy, &dummy, &dummy, &dummy);
		graf_mouse(ARROW, NULL);
		graf_mouse(M_OFF, NULL);		/* turn off mouse */

		draw_bitmap();

		graf_mouse(M_ON, NULL);			/* turn mouse on again */
		appl_exit();
	}

	return 0;
}
